using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Crysanthem
{
    public enum GamePiece
    {
        None,
        None_In_Block,      // Marks the "none" blocks that belong to a piece block
        Player,
        Finish,
        Seed,
        Block,
        Trampoline
    }

    [Flags]
    public enum PieceBorders
    {
        Left = 0x01,
        Right = 0x02,
        Up = 0x04,
        Down = 0x08,
    }

    public class PieceBlock
    {
        public GamePiece[,] pieces;
        public PieceBorders[,] borders;
        public int center_x, center_y;
        public int width, height;

        public int map_x, map_y;

        public PieceBlock(GamePiece[,] _pieces, int w, int h)
        {
            pieces = new GamePiece[w, h];
            borders = new PieceBorders[w, h];
            center_x = w / 2;
            center_y = h / 2;
            width = w;
            height = h;

            map_x = 0;
            map_y = 0;

            for (int y = 0; y < h; ++y)
            {
                for (int x = 0; x < w; ++x)
                {
                    pieces[x, y] = _pieces[x, y];

                    PieceBorders b = 0;
                    if (_pieces[x, y] != GamePiece.None)
                    {
                        if (x <= 0 || _pieces[x - 1, y] == GamePiece.None)
                            b |= PieceBorders.Left;

                        if (x >= (w - 1) || _pieces[x + 1, y] == GamePiece.None)
                            b |= PieceBorders.Right;

                        if (y <= 0 || _pieces[x, y - 1] == GamePiece.None)
                            b |= PieceBorders.Up;

                        if (y >= (h - 1) || _pieces[x, y + 1] == GamePiece.None)
                            b |= PieceBorders.Down;
                    }
                    borders[x, y] = b;
                }
            }
        }

        private PieceBlock()
        {
        }

        public bool IsInBlock(int x, int y)
        {
            x -= map_x;
            x += center_x;
            y -= map_y;
            y += center_y;

            if (x < 0 || y < 0 || x >= width || y >= height)
                return false;

            return
                (pieces[x, y] != GamePiece.None) &&
                (pieces[x, y] != GamePiece.None_In_Block);
        }

        public bool Overlaps(PieceBlock b)
        {
            for (int y = 0; y < height; ++y)
            {
                int my = map_y - center_y + y;
                for (int x = 0; x < width; ++x)
                {
                    if (pieces[x, y] != GamePiece.None)
                    {
                        int mx = map_x - center_x + x;
                        if (b.IsInBlock(mx, my))
                            return true;
                    }
                }
            }

            return false;
        }

        public PieceBlock Copy()
        {
            PieceBlock b = new PieceBlock();

            b.pieces = new GamePiece[width, height];
            b.borders = new PieceBorders[width, height];
            b.center_x = center_x;
            b.center_y = center_y;
            b.width = width;
            b.height = height;

            b.map_x = map_x;
            b.map_y = map_y;

            for (int y = 0; y < height; ++y)
            {
                for (int x = 0; x < width; ++x)
                {
                    b.pieces[x, y] = pieces[x, y];
                    b.borders[x, y] = borders[x, y];
                }
            }

            return b;
        }
    }

    public class GameField
    {
        private const int ScreenWidth = 800;
        private const int ScreenHeight = 600;
        public const int ItemSize = 32;
        public const int ItemWidth = ScreenWidth / ItemSize;
        public const int ItemHeight = ScreenHeight / ItemSize;
        private const int DrawY = ScreenHeight - (ItemHeight * ItemSize);

        private Resources m_res;

        public String sort_name = "";
        public String name = "";
        public int level_index;

        public GamePiece[,] field;
        public PieceBlock[] blocks;

        private float m_visible_angle;

        public int SeedsLeft
        {
            get
            {
                int left = 0;
                for (int y = 0; y < ItemHeight; ++y)
                {
                    for (int x = 0; x < ItemWidth; ++x)
                    {
                        if (field[x, y] == GamePiece.Seed)
                            ++left;
                    }
                }
                return left;
            }
        }

        public GameField(Resources res)
        {
            m_res = res;
            field = new GamePiece[ItemWidth, ItemHeight];
            blocks = new PieceBlock[0];
        }

        public GameField(Resources res, TextReader stm)
            : this(res)
        {
            // Read the name
            name = stm.ReadLine();
            if (name == null)
                name = "";

            // Read the field
            for (int y = 0; y < ItemHeight; ++y)
            {
                String line = stm.ReadLine();

                for (int x = 0; x < ItemWidth; ++x)
                {
                    if (x >= line.Length)
                        break;

                    field[x, y] = ParsePiece(line[x]);
                }
            }

            // Read the piece blocks
            List<PieceBlock> in_blocks = new List<PieceBlock>();
            String next_line = null;

            while (true)
            {
                // Get the next (cached?) line
                String line = next_line;
                next_line = null;
                if (line == null)
                    line = stm.ReadLine();

                // Line ok/recognized?
                if (line == null)
                    break;
                if (line.StartsWith("@PIECE"))
                {
                    // Get the coords
                    int piece_x = 0;
                    int piece_y = 0;
                    String[] args = line.Split(' ');

                    if (args.Length >= 2)
                    {
                        piece_x = Int32.Parse(args[1]);
                        piece_y = Int32.Parse(args[2]);
                    }

                    // Init the read
                    List<List<GamePiece>> in_piece = new List<List<GamePiece>>();
                    int max_w = 0;

                    // Read the list
                    while (true)
                    {
                        // Read a line
                        next_line = stm.ReadLine();
                        if (next_line == null)
                            break;
                        if (next_line.Length > 0 && next_line[0] == '@')
                            break;

                        // Process is
                        List<GamePiece> l = new List<GamePiece>();

                        foreach (char c in next_line)
                            l.Add(ParsePiece(c));
                        in_piece.Add(l);

                        max_w = Math.Max(max_w, next_line.Length);
                    }

                    // Convert the data into a piece list
                    GamePiece[,] piece = new GamePiece[max_w, in_piece.Count];
                    for (int y = 0; y < in_piece.Count; ++y)
                    {
                        List<GamePiece> l = in_piece[y];
                        for (int x = 0; x < l.Count; ++x)
                            piece[x, y] = l[x];
                    }

                    // Add the block
                    if (max_w > 0 && in_piece.Count > 0)
                    {
                        PieceBlock block = new PieceBlock(piece, max_w, in_piece.Count);
                        block.map_x = piece_x + block.center_x;
                        block.map_y = piece_y + block.center_y;

                        in_blocks.Add(block);
                    }
                }
                else if (line.Equals("@END"))
                {
                    break;
                }
                else
                {
                    Console.WriteLine("Warning: unknown directive: '{0}'", line);
                    break;
                }
            }

            blocks = in_blocks.ToArray();
        }

        public GameField Copy()
        {
            GameField f = new GameField(m_res);
            f.sort_name = sort_name;
            f.name = name;
            f.level_index = level_index;

            for (int y = 0; y < ItemHeight; ++y)
                for (int x = 0; x < ItemWidth; ++x)
                    f.field[x, y] = field[x, y];

            f.blocks = new PieceBlock[blocks.Length];
            for (int i = 0; i < blocks.Length; ++i)
                f.blocks[i] = blocks[i].Copy();

            return f;
        }

        public void BakeBlocks()
        {
            foreach (PieceBlock b in blocks)
            {
                for (int y = 0; y < b.height; ++y)
                {
                    int my = b.map_y - b.center_y + y;
                    for (int x = 0; x < b.width; ++x)
                    {
                        int mx = b.map_x - b.center_x + x;

                        if (b.pieces[x, y] != GamePiece.None &&
                            b.pieces[x, y] != GamePiece.None_In_Block)
                        {
                            field[mx, my] = b.pieces[x, y];
                        }
                    }
                }
            }
            blocks = new PieceBlock[0];
        }

        public Vector2 MapToWorld(int map_x, int map_y)
        {
            float x = (map_x * ItemSize) + (ItemSize / 2);
            float y = (map_y * ItemSize) + DrawY + (ItemSize / 2);

            return new Vector2(x, y);
        }

        public bool WorldToMap(float world_x, float world_y, out int map_x, out int map_y)
        {
            world_y -= DrawY;

            map_x = (int)(world_x / (float)ItemSize);
            map_y = (int)(world_y / (float)ItemSize);

            if (world_x < 0 || world_y < 0)
                return false;
            if (map_x >= ItemWidth || map_y >= ItemHeight)
                return false;
            return true;
        }

        public bool IsWorldCoordOk(float world_x, float world_y)
        {
            int x, y;
            return WorldToMap(world_x, world_y, out x, out y);
        }

        public GamePiece GetWorldPiece(float world_x, float world_y)
        {
            int x, y;
            if (!WorldToMap(world_x, world_y, out x, out y))
                return GamePiece.None;
            return field[x, y];
        }

        public void SetWorldPiece(float world_x, float world_y, GamePiece piece)
        {
            int x, y;
            if (WorldToMap(world_x, world_y, out x, out y))
                field[x, y] = piece;
        }

        public bool BlockOverlaps(PieceBlock block)
        {
            if (block == null)
                return false;

            // Check against map
            for (int y = 0; y < block.height; ++y)
            {
                int my = block.map_y - block.center_y + y;
                if (my < 0 || my >= ItemHeight)
                    return true;

                for (int x = 0; x < block.width; ++x)
                {
                    if (block.pieces[x, y] != GamePiece.None &&
                        block.pieces[x, y] != GamePiece.None_In_Block)
                    {
                        int mx = block.map_x - block.center_x + x;
                        if (mx < 0 || mx >= ItemWidth)
                            return true;

                        if (field[mx, my] != GamePiece.None)
                            return true;
                    }
                }
            }

            // Check against other blocks
            for (int i = 0; i < blocks.Length; ++i)
            {
                if (blocks[i] != block && block.Overlaps(blocks[i]))
                    return true;
            }

            return false;
        }

        public PieceBlock GetBlock(int map_x, int map_y)
        {
            for (int i = 0; i < blocks.Length; ++i)
            {
                if (blocks[i].IsInBlock(map_x, map_y))
                    return blocks[i];
            }
            return null;
        }

        public void Update(GameTime time)
        {
            m_visible_angle += (float)time.ElapsedRealTime.TotalSeconds;
            while (m_visible_angle >= (float)Math.PI * 2.0f)
            {
                m_visible_angle -= (float)Math.PI * 2.0f;
            }
        }

        public void Draw(bool draw_edit_tokens)
        {
            Draw(draw_edit_tokens, null);
        }

        public void Draw(bool draw_edit_tokens, PieceBlock editing_block)
        {
            Vector3 pos = new Vector3();

            // Draw the field
            pos.Y = DrawY + ItemSize / 2;
            for (int iy = 0; iy < ItemHeight; ++iy, pos.Y += ItemSize)
            {
                pos.X = ItemSize / 2;
                for (int ix = 0; ix < ItemWidth; ++ix, pos.X += ItemSize)
                {
                    DrawPiece(ref pos, field[ix, iy], draw_edit_tokens);
                }
            }

            // Draw the blocks
            for (int i = 0; i < blocks.Length; ++i)
            {
                Color c = Color.LightGreen;
                if (blocks[i] == editing_block)
                {
                    c = Color.Yellow;

                    if (BlockOverlaps(editing_block))
                        c = Color.Red;
                }

                DrawBlock(blocks[i], c);
            }
        }

        private void DrawPiece(ref Vector3 pos, GamePiece piece, bool draw_edit_tokens)
        {
            Matrix mtx;
            Matrix.CreateTranslation(ref pos, out mtx);

            switch (piece)
            {
                case GamePiece.Player:
                    if (draw_edit_tokens)
                    {
                        mtx = Matrix.CreateRotationZ(m_visible_angle) * mtx;
                        m_res.item_renderer.DrawFlower(ref mtx);
                    }
                    break;

                case GamePiece.Seed:
                    mtx = Matrix.CreateRotationX(MathHelper.ToRadians(15)) * Matrix.CreateRotationY(m_visible_angle) * mtx;
                    m_res.item_renderer.DrawSeed(ref mtx);
                    break;

                case GamePiece.Block:
                    m_res.item_renderer.DrawBlock(ref mtx);
                    break;

                case GamePiece.Trampoline:
                    break;

                case GamePiece.Finish:
                    m_res.item_renderer.DrawFinish(ref mtx);
                    break;
            }
        }

        private Vector2[] m_temp_block_lines = new Vector2[2 * 4];

        public void DrawBlock(PieceBlock draw, Color border_col)
        {
            int map_x = draw.map_x - draw.center_x;
            int map_y = draw.map_y - draw.center_y;

            DrawBlock(draw, map_x, map_y, border_col);
        }

        public void DrawBlock(PieceBlock draw, int map_x, int map_y, Color border_col)
        {
            Vector3 draw_pos = new Vector3(ItemSize / 2, DrawY + ItemSize / 2, 0);
            Vector3 pos = new Vector3();

            int my = map_y;
            for (int y = 0; y < draw.height; ++y, ++my)
            {
                if (my < 0 || my >= ItemHeight)
                    continue;

                int mx = map_x;
                for (int x = 0; x < draw.width; ++x, ++mx)
                {
                    if (mx < 0 || mx >= ItemWidth)
                        continue;

                    pos.X = draw_pos.X + mx * ItemSize;
                    pos.Y = draw_pos.Y + my * ItemSize;

                    // Draw the piece
                    DrawPiece(ref pos, draw.pieces[x, y], true);

                    // Draw the borders
                    const float badd = ItemSize / 2;

                    int bamt = 0;
                    PieceBorders b = draw.borders[x, y];

                    if ((b & PieceBorders.Left) == PieceBorders.Left)
                    {
                        m_temp_block_lines[bamt++] = new Vector2(pos.X - badd, pos.Y - badd);
                        m_temp_block_lines[bamt++] = new Vector2(pos.X - badd, pos.Y + badd);
                    }
                    if ((b & PieceBorders.Right) == PieceBorders.Right)
                    {
                        m_temp_block_lines[bamt++] = new Vector2(pos.X + badd, pos.Y - badd);
                        m_temp_block_lines[bamt++] = new Vector2(pos.X + badd, pos.Y + badd);
                    }
                    if ((b & PieceBorders.Up) == PieceBorders.Up)
                    {
                        m_temp_block_lines[bamt++] = new Vector2(pos.X - badd, pos.Y - badd);
                        m_temp_block_lines[bamt++] = new Vector2(pos.X + badd, pos.Y - badd);
                    }
                    if ((b & PieceBorders.Down) == PieceBorders.Down)
                    {
                        m_temp_block_lines[bamt++] = new Vector2(pos.X - badd, pos.Y + badd);
                        m_temp_block_lines[bamt++] = new Vector2(pos.X + badd, pos.Y + badd);
                    }

                    if (bamt != 0)
                    {
                        m_res.item_renderer.DrawLines(m_temp_block_lines, bamt, false, border_col);
                    }
                }
            }
        }

        private static GamePiece ParsePiece(char c)
        {
            switch (c)
            {
                case 'X':
                    return GamePiece.Block;

                case 'T':
                    return GamePiece.Trampoline;

                case 'S':
                    return GamePiece.Player;

                case 'E':
                    return GamePiece.Finish;

                case '!':
                    return GamePiece.Seed;

                case '_':
                    return GamePiece.None_In_Block;

                default:
                    return GamePiece.None;
            }
        }
    }
}
