using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Storage;
using Crysanthem.Screen;

namespace Crysanthem
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class GameMain : Microsoft.Xna.Framework.Game, IScreenControl
    {
        GraphicsDeviceManager graphics;
        ContentManager content;

        private Resources m_res;

        private GameScreen m_screen_game;
        private EditorScreen m_screen_editor;

        private Viewport m_orig_viewport;
        private Viewport m_new_viewport;

        public GameMain()
        {
            graphics = new GraphicsDeviceManager(this);
            content = new ContentManager(Services);
        }


        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            m_res = new Resources(this, graphics, content);
            m_screen_game = new GameScreen(m_res);
            m_screen_editor = new EditorScreen(m_res);

            Components.Add(m_screen_game);
            Components.Add(m_screen_editor);
            base.Initialize();

            StartEditor(m_res.fields[0]);
        }


        /// <summary>
        /// Load your graphics content.  If loadAllContent is true, you should
        /// load content from both ResourceManagementMode pools.  Otherwise, just
        /// load ResourceManagementMode.Manual content.
        /// </summary>
        /// <param name="loadAllContent">Which type of content to load.</param>
        protected override void LoadGraphicsContent(bool loadAllContent)
        {
            if (loadAllContent)
            {
                m_res.Load();

                m_orig_viewport = graphics.GraphicsDevice.Viewport;
                m_new_viewport = m_orig_viewport;

                // Make a viewport that is a bit smaller
                int w = graphics.GraphicsDevice.DisplayMode.Width;
                int h = graphics.GraphicsDevice.DisplayMode.Height;
                int x = (int)(w * 0.1f);
                int y = (int)(h * 0.1f);
                m_new_viewport.X = x;
                m_new_viewport.Y = y;
                m_new_viewport.Width -= x * 2;
                m_new_viewport.Height -= y * 2;
            }
        }


        /// <summary>
        /// Unload your graphics content.  If unloadAllContent is true, you should
        /// unload content from both ResourceManagementMode pools.  Otherwise, just
        /// unload ResourceManagementMode.Manual content.  Manual content will get
        /// Disposed by the GraphicsDevice during a Reset.
        /// </summary>
        /// <param name="unloadAllContent">Which type of content to unload.</param>
        protected override void UnloadGraphicsContent(bool unloadAllContent)
        {
            if (unloadAllContent)
            {
                m_res.Unload();
                content.Unload();
            }
        }


        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            base.Update(gameTime);
        }


        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
#if XBOX
            // Clear the borders black
            graphics.GraphicsDevice.Viewport = m_orig_viewport;
            graphics.GraphicsDevice.Clear(Color.Black);

            // Make the viewport a bit smaller
            graphics.GraphicsDevice.Viewport = m_new_viewport;
            graphics.GraphicsDevice.Clear(Color.CornflowerBlue);
#else
            graphics.GraphicsDevice.Clear(Color.CornflowerBlue);
#endif

            base.Draw(gameTime);
        }

        private void HideScreens()
        {
            m_screen_game.Enabled = false;
            m_screen_game.Visible = false;

            m_screen_editor.Enabled = false;
            m_screen_editor.Visible = false;
        }

        public void StartGame(GameField field)
        {
            HideScreens();
            m_screen_game.Enabled = true;
            m_screen_game.Visible = true;

            m_screen_game.Init(field);
        }

        public void StartEditor(GameField field)
        {
            HideScreens();
            m_screen_editor.Enabled = true;
            m_screen_editor.Visible = true;

            m_screen_editor.Init(field);
        }

        public void BackToEditor()
        {
            HideScreens();
            m_screen_editor.Enabled = true;
            m_screen_editor.Visible = true;
        }

        public void StartLevelSelect()
        {
            HideScreens();
        }
    }
}
