using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Crysanthem
{
    public class ItemRenderer
    {
        private const float ItemWidth = 32.0f;
        private const float ItemRadius = ItemWidth * 0.5f;

        private BasicEffect m_effect;
        private BasicEffect m_effect_lines;
        private GraphicsDevice m_device;

        private VertexDeclaration m_vertex_declaration;
        private VertexDeclaration m_vertex_declaration_lines;

        private VertexPositionNormalColor[] m_vtx_flower_petal;
        private VertexPositionNormalColor[] m_vtx_flower_center;
        private VertexPositionNormalColor[] m_vtx_finish_petal;
        private VertexPositionNormalColor[] m_vtx_seed;
        private VertexPositionNormalColor[] m_vtx_block;
        private VertexPositionColor[] m_vtx_cursor;

        private VertexPositionColor[] m_vtx_line_cache;

        public ItemRenderer(GraphicsDevice device)
        {
            m_device = device;

            m_vertex_declaration = new VertexDeclaration(m_device, VertexPositionNormalColor.VertexElements);
            m_vertex_declaration_lines = new VertexDeclaration(m_device, VertexPositionColor.VertexElements);

            // Create the basic effect
            m_effect = new BasicEffect(device, null);
            m_effect.Alpha = 1.0f;
            m_effect.DiffuseColor = new Vector3(1.0f, 1.0f, 1.0f);
            m_effect.SpecularColor = new Vector3(0.25f, 0.25f, 0.25f);
            m_effect.SpecularPower = 5.0f;
            m_effect.AmbientLightColor = new Vector3(0.15f, 0.15f, 0.15f);

            m_effect.DirectionalLight0.Enabled = true;
            m_effect.DirectionalLight0.DiffuseColor = Vector3.One;
            m_effect.DirectionalLight0.Direction = Vector3.Normalize(new Vector3(1.0f, -1.0f, -1.0f));
            m_effect.DirectionalLight0.SpecularColor = Vector3.One;

            m_effect.LightingEnabled = true;
            m_effect.VertexColorEnabled = true;
            m_effect.TextureEnabled = false;

            m_effect.View = Matrix.CreateLookAt(new Vector3(0, 0, 50), new Vector3(0, 0, 0), new Vector3(0, 1, 0));
            m_effect.Projection = Matrix.CreateOrthographicOffCenter(0, 800, 600, 0, 0.1f, 100.0f);

            // Create the line effect
            m_effect_lines = new BasicEffect(device, null);
            m_effect_lines.Alpha = 1.0f;
            m_effect_lines.DiffuseColor = new Vector3(1.0f, 1.0f, 1.0f);
            m_effect_lines.AmbientLightColor = new Vector3(0.15f, 0.15f, 0.15f);

            m_effect_lines.LightingEnabled = false;
            m_effect_lines.VertexColorEnabled = true;
            m_effect_lines.TextureEnabled = false;

            m_effect_lines.View = Matrix.CreateLookAt(new Vector3(0, 0, 50), new Vector3(0, 0, 0), new Vector3(0, 1, 0));
            m_effect_lines.Projection = Matrix.CreateOrthographicOffCenter(0, 800, 600, 0, 0.1f, 100.0f);

            // Create the meshes
            CreateFlowerCenter();
            m_vtx_flower_petal = CreatePetal(Color.Yellow);
            m_vtx_finish_petal = CreatePetal(Color.Green);
            CreateSeed();
            CreateBlock();
            CreateCursor();
        }

        public void DrawFlower(ref Matrix mtx)
        {
            m_device.VertexDeclaration = m_vertex_declaration;

            const int petal_count = 10;

            Matrix tmp_mtx1, tmp_mtx2;

            m_effect.Begin();
            foreach (EffectPass pass in m_effect.CurrentTechnique.Passes)
            {
                pass.Begin();

                // Draw the petals
                for (int i = 0; i < petal_count; ++i)
                {
                    float angle = (float)Math.PI * 2.0f * ((float)i / (float)petal_count);

                    Matrix.CreateRotationZ(angle, out tmp_mtx1);
                    Matrix.Multiply(ref tmp_mtx1, ref mtx, out tmp_mtx2);
                    m_effect.World = tmp_mtx2;
                    m_effect.CommitChanges();

                    m_device.DrawUserPrimitives<VertexPositionNormalColor>(PrimitiveType.TriangleFan, m_vtx_flower_petal, 0, m_vtx_flower_petal.Length - 2);
                }

                // Draw the center
                m_effect.World = mtx;
                m_effect.CommitChanges();
                m_device.DrawUserPrimitives<VertexPositionNormalColor>(PrimitiveType.TriangleFan, m_vtx_flower_center, 0, m_vtx_flower_center.Length - 2);

                pass.End();
            }
            m_effect.End();
        }

        public void DrawFinish(ref Matrix mtx)
        {
            m_device.VertexDeclaration = m_vertex_declaration;

            const int petal_count = 5;
            const float move_per_petal = 5.0f;

            float tilt_angle = MathHelper.ToRadians(40.0f);

            Matrix tmp_mtx1, tmp_mtx2;

            m_effect.Begin();
            foreach (EffectPass pass in m_effect.CurrentTechnique.Passes)
            {
                pass.Begin();

                // Draw the petals
                float y = ItemRadius;
                for (int i = 0; i < petal_count; ++i, y -= move_per_petal)
                {
                    float angle = tilt_angle;
                    if ((i & 1) == 1) angle = -angle;

                    Matrix.CreateRotationZ(angle, out tmp_mtx1);
                    tmp_mtx1.M42 = y;
                    Matrix.Multiply(ref tmp_mtx1, ref mtx, out tmp_mtx2);
                    m_effect.World = tmp_mtx2;
                    m_effect.CommitChanges();

                    m_device.DrawUserPrimitives<VertexPositionNormalColor>(PrimitiveType.TriangleFan, m_vtx_flower_petal, 0, m_vtx_flower_petal.Length - 2);
                }

                pass.End();
            }
            m_effect.End();
        }

        public void DrawSeed(ref Matrix mtx)
        {
            m_device.VertexDeclaration = m_vertex_declaration;

            m_effect.World = mtx;

            m_effect.Begin();
            foreach (EffectPass pass in m_effect.CurrentTechnique.Passes)
            {
                pass.Begin();
                m_device.DrawUserPrimitives<VertexPositionNormalColor>(PrimitiveType.TriangleStrip, m_vtx_seed, 0, m_vtx_seed.Length - 2);
                pass.End();
            }
            m_effect.End();
        }

        public void DrawBlock(ref Matrix mtx)
        {
            m_device.VertexDeclaration = m_vertex_declaration;

            m_effect.World = mtx;

            m_effect.Begin();
            foreach (EffectPass pass in m_effect.CurrentTechnique.Passes)
            {
                pass.Begin();
                m_device.DrawUserPrimitives<VertexPositionNormalColor>(PrimitiveType.TriangleFan, m_vtx_block, 0, m_vtx_block.Length - 2);
                pass.End();
            }
            m_effect.End();
        }

        public void DrawCursor(Vector2 pos)
        {
            Matrix mtx;
            Matrix.CreateTranslation(pos.X, pos.Y, 0, out mtx);

            DrawLines(ref mtx, PrimitiveType.LineList, m_vtx_cursor, m_vtx_cursor.Length);
        }

        public void DrawLines(Vector2[] line_data, int point_count, bool is_strip, Color color)
        {
            int amt = point_count;
            if (m_vtx_line_cache == null || m_vtx_line_cache.Length < amt)
            {
                m_vtx_line_cache = new VertexPositionColor[amt];
            }

            for (int i = 0; i < amt; ++i)
            {
                m_vtx_line_cache[i].Position = new Vector3(line_data[i].X, line_data[i].Y, 5.0f);
                m_vtx_line_cache[i].Color = color;
            }

            Matrix mtx = Matrix.Identity;
            DrawLines(ref mtx, is_strip ? PrimitiveType.LineStrip : PrimitiveType.LineList, m_vtx_line_cache, amt);
        }

        private void DrawLines(ref Matrix mtx, PrimitiveType prim_type, VertexPositionColor[] line_points, int point_count)
        {
            if (line_points == null || point_count <= 1)
                return;

            m_device.VertexDeclaration = m_vertex_declaration_lines;

            m_effect_lines.World = mtx;

            m_effect_lines.Begin();
            foreach (EffectPass pass in m_effect_lines.CurrentTechnique.Passes)
            {
                pass.Begin();

                int amt = point_count;
                if (prim_type == PrimitiveType.LineStrip)
                    amt -= 1;
                else
                    amt /= 2;

                m_device.DrawUserPrimitives<VertexPositionColor>(prim_type, line_points, 0, amt);
                pass.End();
            }
            m_effect_lines.End();
        }

        private VertexPositionNormalColor[] CreatePetal(Color color)
        {
            List<VertexPositionNormalColor> vtx = new List<VertexPositionNormalColor>();

            const float petal_radius = ItemRadius;
            const float center_y = petal_radius * 0.75f;

            // Add the center point
            vtx.Add(new VertexPositionNormalColor(new Vector3(0, -center_y, 1), new Vector3(0, 0, 1), color));

            // Add the corner points
            for (int ang_idx = 0; ang_idx <= 32; ++ang_idx)
            {
                float angle = -(float)Math.PI*0.5f + (float)Math.PI * 2.0f * ((float)ang_idx / 32.0f);

                float dir_x = (float)Math.Cos(angle);
                float dir_y = (float)Math.Sin(angle);

                float y_radius = MathUtil.MapRange(dir_y, -1, 1, petal_radius - center_y, center_y);
                float x_radius = 3.0f;

                Vector3 pos = new Vector3(dir_x * x_radius, -center_y + dir_y * y_radius, 0);
                Vector3 normal = Vector3.Normalize(new Vector3(dir_x, dir_y, -0.3f));

                vtx.Add(new VertexPositionNormalColor(pos, normal, color));
            }

            return vtx.ToArray();

        }

        private void CreateFlowerCenter()
        {
            List<VertexPositionNormalColor> vtx = new List<VertexPositionNormalColor>();

            const float radius = ItemRadius * 0.4f;
            const int points = 32;

            // Add the center point
            vtx.Add(new VertexPositionNormalColor(new Vector3(0, 0, 2), new Vector3(0, 0, 1), Color.Yellow));

            // Add the corner points
            for (int ang_idx = 0; ang_idx <= points; ++ang_idx)
            {
                float angle = (float)Math.PI * 2.0f * ((float)ang_idx / (float)points);

                float dir_x = (float)Math.Cos(angle);
                float dir_y = (float)Math.Sin(angle);

                float wave_angle = angle * 10.0f;
                float real_radius = radius * MathUtil.MapRange((float)Math.Cos(wave_angle), -1, 1, 0.9f, 1.1f);

                Vector3 pos = new Vector3(dir_x * real_radius, dir_y * real_radius, 1);
                Vector3 normal = Vector3.Normalize(pos);

                vtx.Add(new VertexPositionNormalColor(pos, normal, Color.Yellow));
            }

            m_vtx_flower_center = vtx.ToArray();
        }

        private void CreateSeed()
        {
            List<VertexPositionNormalColor> vtx = new List<VertexPositionNormalColor>();

            const int sections = 5;
            const int section_parts = 6;
            const int vert_parts = 12;

            const float base_radius = ItemRadius * 0.7f;
            const float shadow_radius = base_radius * 1.25f;

            for (int section = 0; section < sections; ++section)
            {
                float section_angle1 = (float)Math.PI * 2.0f * ((float)section / sections);
                float section_angle2 = (float)Math.PI * 2.0f * ((float)(section + 1) / sections);
                float section_angle_center = (section_angle1 + section_angle2) * 0.5f;
                float max_dist = section_angle2 - section_angle_center;

                for (int section_part = 0; section_part < section_parts; ++section_part)
                {
                    float horiz_angle1 = MathUtil.MapRange((float)section_part, 0, section_parts, section_angle1, section_angle2);
                    float horiz_angle2 = MathUtil.MapRange((float)(section_part + 1), 0, section_parts, section_angle1, section_angle2);

                    float radius1 = MathUtil.MapRange(Math.Abs(horiz_angle1 - section_angle_center), 0.0f, max_dist, 0.0f, (float)Math.PI);
                    radius1 = base_radius * (1.0f + (float)Math.Sin(radius1) * 0.55f);

                    float radius2 = MathUtil.MapRange(Math.Abs(horiz_angle2 - section_angle_center), 0.0f, max_dist, 0.0f, (float)Math.PI);
                    radius2 = base_radius * (1.0f + (float)Math.Sin(radius2) * 0.55f);

                    // (one loop)
                    for (int vert_part = 0; vert_part <= (vert_parts * 2); ++vert_part)
                    {
                        float vert_angle = (float)Math.PI * ((float)vert_part / vert_parts);

                        float mul = Math.Abs((float)Math.Sin(vert_angle));
                        float y = (float)Math.Cos(vert_angle);

                        mul *= MathUtil.MapRange(y, -1.0f, 1.0f, 0.0f, 1.0f);

                        float x1 = (float)Math.Cos(horiz_angle1) * mul;
                        float y1 = y;
                        float z1 = (float)Math.Sin(horiz_angle1) * mul;

                        float x2 = (float)Math.Cos(horiz_angle2) * mul;
                        float y2 = y;
                        float z2 = (float)Math.Sin(horiz_angle2) * mul;

                        float normal_x = (float)Math.Cos(section_angle_center) * mul;
                        float normal_y = y;
                        float normal_z = (float)Math.Sin(section_angle_center) * mul;
                        Vector3 normal_pos = new Vector3(normal_x, normal_y, normal_z) * shadow_radius;

                        Vector3 pos1 = new Vector3(x1, y1, z1) * radius1;
                        Vector3 normal1 = Vector3.Normalize(pos1 - normal_pos);

                        Vector3 pos2 = new Vector3(x2, y2, z2) * radius2;
                        Vector3 normal2 = Vector3.Normalize(pos2 - normal_pos);

                        vtx.Add(new VertexPositionNormalColor(pos1, normal1, Color.Brown));
                        vtx.Add(new VertexPositionNormalColor(pos2, normal2, Color.Brown));
                    }
                }
            }

            m_vtx_seed = vtx.ToArray();
        }

        private void CreateBlock()
        {
            List<VertexPositionNormalColor> vtx = new List<VertexPositionNormalColor>();

            const float cx = ItemRadius * -0.6f;
            const float cy = ItemRadius * -0.6f;
            const float z = -5.0f;

            // Add the center point
            vtx.Add(new VertexPositionNormalColor(new Vector3(cx, cy, 0), new Vector3(0, 0, 1), Color.Thistle));

            for (int i = 0; i <= 8; ++i)
            {
                float x = -ItemRadius + ItemWidth * ((float)i / 8);

                Vector3 pos = new Vector3(x, -ItemRadius, z);
                Vector3 norm = Vector3.Normalize(pos);

                vtx.Add(new VertexPositionNormalColor(pos, norm, Color.Thistle));
            }

            for (int i = 8; i >= 0; --i)
            {
                float y = -ItemRadius + ItemWidth * ((float)i / 8);

                Vector3 pos = new Vector3(-ItemRadius, y, z);
                Vector3 norm = Vector3.Normalize(pos);

                vtx.Add(new VertexPositionNormalColor(pos, norm, Color.Thistle));
            }

            for (int i = 8; i >= 0; --i)
            {
                float x = -ItemRadius + ItemWidth * ((float)i / 8);

                Vector3 pos = new Vector3(x, ItemRadius, z);
                Vector3 norm = Vector3.Normalize(pos);

                vtx.Add(new VertexPositionNormalColor(pos, norm, Color.Thistle));
            }

            for (int i = 0; i <= 8; ++i)
            {
                float y = -ItemRadius + ItemWidth * ((float)i / 8);

                Vector3 pos = new Vector3(ItemRadius, y, z);
                Vector3 norm = Vector3.Normalize(pos);

                vtx.Add(new VertexPositionNormalColor(pos, norm, Color.Thistle));
            }

            m_vtx_block = vtx.ToArray();
        }

        private void CreateCursor()
        {
            List<VertexPositionColor> vtx = new List<VertexPositionColor>();

            // The cross
            vtx.Add(new VertexPositionColor(new Vector3(-ItemRadius, 0, 10.0f), Color.WhiteSmoke));
            vtx.Add(new VertexPositionColor(new Vector3(ItemRadius, 0, 10.0f), Color.WhiteSmoke));

            vtx.Add(new VertexPositionColor(new Vector3(0, -ItemRadius, 10.0f), Color.WhiteSmoke));
            vtx.Add(new VertexPositionColor(new Vector3(0, ItemRadius, 10.0f), Color.WhiteSmoke));

            // The circle
            float rad = ItemRadius * 0.5f;

            float step = ((float)Math.PI * 2.0f) / 32.0f;
            for (int i = 0; i < 32; ++i)
            {
                float angle1 = step * (float)i;
                float angle2 = angle1 + step;

                float x1 = (float)Math.Cos(angle1) * rad;
                float y1 = (float)Math.Sin(angle1) * rad;
                float x2 = (float)Math.Cos(angle2) * rad;
                float y2 = (float)Math.Sin(angle2) * rad;

                vtx.Add(new VertexPositionColor(new Vector3(x1, y1, 10.0f), Color.WhiteSmoke));
                vtx.Add(new VertexPositionColor(new Vector3(x2, y2, 10.0f), Color.WhiteSmoke));
            }

            m_vtx_cursor = vtx.ToArray();
        }
    }
}
