using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Crysanthem;

namespace Crysanthem.Screen
{
    public class EditorScreen : Screen
    {
        private Cursor m_cursor;
        private GameField m_field;

        private PieceBlock m_moving_block;
        private bool m_select_down;
        private bool m_next_level_down;
        private bool m_prev_level_down;

        protected override GameField CurrentField
        {
            get { return m_field; }
        }

        protected override String[] InfoLines
        {
            get
            {
                return new String[] {
                    "Crysanthem -- A game to 9th LD48h game competition",
                    "29.-30.4.2007  Jere \"XMunkki\" Sanisalo",
#if !XBOX
                    "Use the mouse to pick hilighted pieces up and place them.",
                    "Press space to play the game.",
                    "Use the arrow keys to change levels.",
                    "You may also use the game pad.",
#else
                    "Use the stick to move around.",
                    "Use the A button to pick hilighted pieces up and place them.",
                    "Use the start button to play the game.",
                    "Use the shoulder buttons to switch levels.",
#endif
                };
            }
        }

        public EditorScreen(Resources res)
            : base(res)
        {
            m_cursor = new Cursor();
        }

        public void Init(GameField field)
        {
            m_cursor.Reset();
            m_field = field.Copy();
            m_moving_block = null;
        }

        public override void Update(GameTime gameTime)
        {
            base.Update(gameTime);

            m_cursor.Update();
            m_field.Update(gameTime);

            // Get the cursor position
            int cursor_x, cursor_y;
            bool cursor_ok = m_field.WorldToMap(m_cursor.Position.X, m_cursor.Position.Y, out cursor_x, out cursor_y);

            // Controls
            bool starting = false;
            bool select_down = false;
            bool next_level_down = false;
            bool prev_level_down = false;

#if !XBOX
            if (Keyboard.GetState().IsKeyDown(Keys.Space))
                starting = true;
            if (Keyboard.GetState().IsKeyDown(Keys.Enter))
                select_down = true;
            if (Keyboard.GetState().IsKeyDown(Keys.Left))
                prev_level_down = true;
            if (Keyboard.GetState().IsKeyDown(Keys.Right))
                next_level_down = true;
            if (Mouse.GetState().LeftButton == ButtonState.Pressed)
                select_down = true;
#endif
            if (GamePad.GetState(PlayerIndex.One).Buttons.A == ButtonState.Pressed)
                select_down = true;
            if (GamePad.GetState(PlayerIndex.One).Buttons.Start == ButtonState.Pressed)
                starting = true;
            if (GamePad.GetState(PlayerIndex.One).Buttons.LeftShoulder == ButtonState.Pressed)
                prev_level_down = true;
            if (GamePad.GetState(PlayerIndex.One).Buttons.RightShoulder == ButtonState.Pressed)
                next_level_down = true;

            if (select_down && !m_select_down)
            { // Select pressed
                m_select_down = true;
            }
            if (!select_down && m_select_down)
            { // Select lifted
                m_select_down = false;

                if (m_moving_block != null)
                { // Placing a block
                    if (cursor_ok && m_field.BlockOverlaps(m_moving_block) == false)
                        m_moving_block = null;
                }
                else
                { // Selecting a block from field / unused block
                    if (cursor_ok)
                        m_moving_block = m_field.GetBlock(cursor_x, cursor_y);
                }
            }

            if (next_level_down)
                m_next_level_down = true;
            else if (m_next_level_down)
            {
                m_next_level_down = false;
                if (!starting && !IsAnyKeyDown() && m_moving_block == null)
                {
                    ChangeLevel(1);
                }
            }

            if (prev_level_down)
                m_prev_level_down = true;
            else if (m_prev_level_down)
            {
                m_prev_level_down = false;
                if (!starting && !IsAnyKeyDown() && m_moving_block == null)
                {
                    ChangeLevel(-1);
                }
            }

            if (starting &&
                !IsAnyKeyDown() &&
                m_moving_block == null)
            {
                res.screen_control.StartGame(m_field);
            }

            // Moving a block?
            if (m_moving_block != null && cursor_ok)
            {
                m_moving_block.map_x = cursor_x;
                m_moving_block.map_y = cursor_y;
            }
        }

        private bool IsAnyKeyDown()
        {
            return m_select_down || m_next_level_down || m_prev_level_down;
        }

        private void ChangeLevel(int dir)
        {
            int idx = m_field.level_index + dir;
            if (idx < 0)
                idx = res.fields.Count - 1;
            if (idx >= res.fields.Count)
                idx = 0;

            Init(res.fields[idx]);
        }


        public override void Draw(GameTime gameTime)
        {
            base.Draw(gameTime);

            m_field.Draw(true, m_moving_block);

            // Draw the cursor
            if (m_moving_block == null)
                res.item_renderer.DrawCursor(m_cursor.Position);

            // Show the time
            float top_time = res.fields.GetTime(m_field.level_index);
            String time_txt = "Top time: ";
            if (top_time >= 0.0f)
                time_txt += MathUtil.FloatToString(top_time);
            else
                time_txt += "NOT SET";

            res.sprite_batch.Begin();
            res.sprite_batch.DrawString(
                res.font,
                time_txt,
                new Vector2(4.0f, 16.0f),
                Color.Black);
            res.sprite_batch.End();
        }
    }
}
