using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Crysanthem;

namespace Crysanthem.Screen
{
    public class GameScreen : Screen
    {
        private static Vector2 Gravity = new Vector2(0, 400.0f);
        private const float MaxVelocity = 300.0f;
        private const float Acceleration = 200.0f;
        private const float JumpVelocity = 250.0f;
        private const float FrictionMul = 0.98f;

        private GameField m_original_field;
        private GameField m_field;

        private Vector2 m_pos;
        private Vector2 m_vel;
        private float m_time;
        private float m_angle;

        private bool m_finished;
        private bool m_exit_down;

        private Vector2 m_finish_pos;

        protected override GameField CurrentField
        {
            get { return m_field; }
        }

        protected override String[] InfoLines
        {
            get
            {
                return new String[] {
#if !XBOX
                    "Use the arrow keys to move and jump.",
                    "Collect all seeds and get to the exit.",
                    "Press space to restart the level.",
                    "Press the escape button to return to the editor.",
#else
                    "Use the left stick to move and the A button jump.",
                    "Collect all seeds and get to the exit.",
                    "Press the start button to restart the level.",
                    "Press the back button to return to the editor.",
#endif
                };
            }
        }

        public GameScreen(Resources res)
            : base(res)
        {
        }

        public void Init(GameField field)
        {
            if (field != m_original_field)
            {
                m_original_field = field.Copy();
                m_original_field.BakeBlocks();
            }
            m_field = m_original_field.Copy();

            // Find the starting position
            m_pos = Vector2.Zero;
            m_vel = Vector2.Zero;
            m_time = 0.0f;
            m_angle = 0.0f;

            m_finished = false;

            m_finish_pos = Vector2.Zero;

            for (int y = 0; y < GameField.ItemHeight; ++y)
            {
                for (int x = 0; x < GameField.ItemWidth; ++x)
                {
                    if (m_field.field[x, y] == GamePiece.Player)
                    {
                        m_pos = m_field.MapToWorld(x, y);
                    }
                    if (m_field.field[x, y] == GamePiece.Finish)
                    {
                        m_finish_pos = m_field.MapToWorld(x, y);
                    }
                }
            }
        }

        public override void Update(GameTime gameTime)
        {
            base.Update(gameTime);

            m_field.Update(gameTime);

            float delta_t = (float)gameTime.ElapsedRealTime.TotalSeconds;

            bool running = !m_finished;

            if (running)
            {
                // Update the running time
                m_time += delta_t;

                // Update the velocity
                m_vel += Gravity * delta_t;

                if (m_vel.Length() > MaxVelocity)
                { // Clamp to max velocity
                    m_vel.Normalize();
                    m_vel *= MaxVelocity;
                }

                // Update the position
                m_pos += m_vel * delta_t;
            }
            else
            {
                Vector2 dest = m_finish_pos;
                dest.Y -= 15.0f;

                m_pos = Vector2.Lerp(m_pos, dest, 0.05f);
            }

            // Collision detection
            const float collision_radius = (float)GameField.ItemSize * 0.5f;

            GamePiece piece_pos = m_field.GetWorldPiece(m_pos.X, m_pos.Y);
            int cur_map_x, cur_map_y;

            bool on_ground = false;
            bool out_of_map = !m_field.WorldToMap(m_pos.X, m_pos.Y, out cur_map_x, out cur_map_y);

            if (!out_of_map &&
                !PositionBlocked(m_pos.X, m_pos.Y) &&
                running)
            {
                Vector2 map_coord = m_field.MapToWorld(cur_map_x, cur_map_y);

                if (TestForCollision(m_pos.X - collision_radius, m_pos.Y) ||
                    TestForCollision(m_pos.X + collision_radius, m_pos.Y))
                {
                    m_pos.X = map_coord.X;
                    m_vel.X = 0.0f;
                }

                on_ground = TestForCollision(m_pos.X, m_pos.Y + collision_radius);
                if (TestForCollision(m_pos.X, m_pos.Y - collision_radius) ||
                    on_ground)
                {
                    m_pos.Y = map_coord.Y;
                    m_vel.Y = 0.0f;
                }
            }

            // On ground friction?
            if (on_ground && running)
                m_vel.X *= FrictionMul;

            // Update the rotation angle
            if (running)
                m_angle += m_vel.X * 0.001f;
            if (m_finished)
                m_angle += 0.02f;

            while (m_angle < 0.0f)
                m_angle += (float)Math.PI * 2.0f;

            while (m_angle >= (float)Math.PI * 2.0f)
                m_angle -= (float)Math.PI * 2.0f;

            // Controls
            int x_acc = 0;
            bool jumping = false;
            bool exiting = false;
            bool restart = false;

#if !XBOX
            if (Keyboard.GetState().IsKeyDown(Keys.Left))
                x_acc -= 1;
            if (Keyboard.GetState().IsKeyDown(Keys.Right))
                x_acc += 1;
            if (Keyboard.GetState().IsKeyDown(Keys.Up))
                jumping = true;
            if (Keyboard.GetState().IsKeyDown(Keys.Escape))
                exiting = true;
            if (Keyboard.GetState().IsKeyDown(Keys.Space))
                restart = true;
#endif

            if (GamePad.GetState(PlayerIndex.One).ThumbSticks.Left.X < -0.5f)
                x_acc -= 1;
            if (GamePad.GetState(PlayerIndex.One).ThumbSticks.Left.X > 0.5f)
                x_acc += 1;
            if (GamePad.GetState(PlayerIndex.One).Buttons.A == ButtonState.Pressed)
                jumping = true;
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                exiting = true;
            if (GamePad.GetState(PlayerIndex.One).Buttons.Start == ButtonState.Pressed)
                restart = true;

            if (running)
            {
                m_vel.X += (float)x_acc * Acceleration * delta_t;
                if (on_ground && jumping)
                    m_vel.Y = -JumpVelocity;

                // Out of map?
                if (out_of_map)
                {
                    Init(m_original_field);
                }
            }

            // Exiting?
            if (exiting)
                m_exit_down = true;
            else if (!exiting && m_exit_down)
            {
                m_exit_down = false;
                res.screen_control.BackToEditor();
            }

            // Restart?
            if (restart)
                Init(m_original_field);
        }

        private bool PositionBlocked(float x, float y)
        {
            switch (m_field.GetWorldPiece(x, y))
            {
                case GamePiece.Block:
                case GamePiece.Trampoline:
                    return true;
            }
            return false;
        }

        private bool TestForCollision(float test_x, float test_y)
        {
            GamePiece p = m_field.GetWorldPiece(test_x, test_y);
            switch (p)
            {
                case GamePiece.Block:
                    return true;

                case GamePiece.Seed:
                    m_field.SetWorldPiece(test_x, test_y, GamePiece.None);
                    break;

                case GamePiece.Finish:
                    if (!m_finished && m_field.SeedsLeft == 0)
                    {
                        m_finished = true;

                        // Check for top time
                        float top_time = res.fields.GetTime(m_field.level_index);
                        if (top_time < 0.0f || m_time < top_time)
                        {
                            res.fields.SetTime(m_field.level_index, m_time);
                        }
                    }
                    break;
            }
            return false;
        }

        public override void Draw(GameTime gameTime)
        {
            Matrix mtx1, mtx2, mtx3;

            base.Draw(gameTime);

            // Draw the field
            m_field.Draw(false);

            // Draw the player
            Matrix.CreateRotationZ(m_angle, out mtx1);
            Matrix.CreateTranslation(m_pos.X, m_pos.Y, 0.0f, out mtx2);
            Matrix.Multiply(ref mtx1, ref mtx2, out mtx3);
            res.item_renderer.DrawFlower(ref mtx3);

            // Show the time
            float top_time = res.fields.GetTime(m_field.level_index);
            String time_txt = "Time: " + MathUtil.FloatToString(m_time);
            if (top_time >= 0.0f)
                time_txt += " / " + MathUtil.FloatToString(top_time);

            res.sprite_batch.Begin();
            res.sprite_batch.DrawString(
                res.font,
                time_txt,
                new Vector2(4.0f, 16.0f),
                Color.Black);

            // Show the amount of seeds left
            res.sprite_batch.DrawString(
                res.font,
                "Seeds left: " + m_field.SeedsLeft,
                new Vector2(4.0f, 32.0f),
                Color.Black);

            res.sprite_batch.End();
        }
    }
}
